
#ifndef __NETFINDER_H__
#define __NETFINDER_H__


#define NETFINDER_SERVER_PORT  3040

// nf_header.magic value
#define NF_MAGIC                0x5A

// nf_header.id values
#define NF_IDENTIFY                     0
#define NF_IDENTIFY_REPLY               1
#define NF_ASSIGNMENT                   2
#define NF_ASSIGNMENT_REPLY             3

// nf_identify_reply.mode values
#define NF_MODE_BOOTLOADER      0   // Bootloader is active
#define NF_MODE_APPLICATION     1   // Application is active

// nf_identify_reply.ip_type and
// nf_assignment.ip_type values
#define NF_IP_DYNAMIC     0
#define NF_IP_STATIC      1

// nf_identify_reply.alert values
#define NF_ALERT_OK             0x00
#define NF_ALERT_WARN           0x01
#define NF_ALERT_ERROR          0xFF

// nf_assignment_reply.result values
#define NF_SUCCESS              0
#define NF_INVALID_IP_TYPE      4


struct nf_header
{
    uint8_t magic;
    uint8_t id;
    uint16_t sequence;
    uint8_t eth_addr[6];
    uint8_t reserved[2];        // Set to 0
} __attribute__ ((__packed__));


struct nf_identify
{
    struct nf_header hdr;
} __attribute__ ((__packed__));


struct nf_identify_reply
{
    struct nf_header hdr;
    uint16_t uptime_days;
    uint8_t uptime_hrs;
    uint8_t uptime_mins;
    uint8_t uptime_secs;
    uint8_t mode;
    uint8_t alert;
    uint8_t ip_type;
    uint8_t ip_addr[4];
    uint8_t ip_netmask[4];
    uint8_t ip_gw[4];
    uint8_t app_ver[4];
    uint8_t boot_ver[4];
    uint8_t hw_ver[4];
    uint8_t name[32];       // Currently not used
} __attribute__ ((__packed__));

struct nf_assignment
{
    struct nf_header hdr;
    uint8_t reserved[3];    // Set to 0
    uint8_t ip_type;
    uint8_t ip_addr[4];
    uint8_t ip_netmask[4];
    uint8_t ip_gw[4];
    uint8_t name[32];   // Currently not used
} __attribute__ ((__packed__));

struct nf_assignment_reply
{
    struct nf_header hdr;
    uint8_t result;
    uint8_t reserved[3];
} __attribute__ ((__packed__));


#endif /* __NETFINDER_H__ */

