/*
 * Prologic USB-ETHERNET CONTROLLER Netwaork Configurator
 *
 * (C) 2009 Kouichi Hirabayashi (kh@mogami-wire.co.jp)
 *
 * Absolutely no warranty. Use this software with your own risk.
 *
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright and disclaimer notice.
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h> 
#include <unistd.h> 
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/sockio.h>
#include <net/if.h>
#include <net/bpf.h>
#include <netdb.h>
#include <ifaddrs.h>
#include <net/if_dl.h>
#include "pgconf.h"

#define TIMEOUT		2

#define PORT_NO		NETFINDER_SERVER_PORT

void onalarm(int);
char *addr(char *);
char *findif(void);
unsigned short ip_checksum(char *s, int n);
unsigned short udp_checksum(char *s, int n);
unsigned int eth_checksum(char *s, int n);

extern int errno;
struct sigaction saalrm, osaalrm;

struct nf_identify nf_identify;
struct nf_assignment nf_assignment;

struct eth_frame eth_frame;	/* Ethernet frame */
struct ip ip;			/* IP packet */
struct udp_header udp_hdr;	/* UDP header */

char *ifname;			/* network interface */
static unsigned char if_mac[6];	/* network interface MAC address */

int bpf;	/* bpf device */
int ufd;	/* UDP socket */

/* GPIB-ETHERNET CONTROLLER Network parameters */
static unsigned char mac_addr[6];
static unsigned char ip_addr[4];
static unsigned char ip_netmask[4];
static unsigned char ip_gw[4];
static struct sockaddr_in sv_addr, cl_addr;

extern int optind;
extern char *optarg;

int vflg;	/* vebose */
char *cmd;
char buf[BUFSIZ];

int
main(int argc, char *argv[])
{
  int ch, on;

  cmd = argv[0];
  while ((ch = getopt(argc, argv, "ihv?")) != -1)
	switch((char)ch) {
	case 'i':
		ifname = optarg;
		break;
	case 'v':
		vflg++;
		break;
	default:
		Usage();
	}
  argc -= optind;
  argv += optind;

  if (argc >= 3) {
	set_arg(argv[0], &ip_addr);
	set_arg(argv[1], &ip_netmask);
	set_arg(argv[2], &ip_gw);
  }
  if (argc == 4)
	set_macaddr(argv[3]);
  else if (1 < argc && argc < 3)
	Usage();
  if (geteuid() != 0)
	error("You must be a super user.", (char *)0);
  saalrm.sa_handler = onalarm;
  sigemptyset(&saalrm.sa_mask);
  saalrm.sa_flags = 0;
  sigaction(SIGALRM, &saalrm, &osaalrm);

  if (ifname == NULL)
	ifname = findif();
  init(ifname);

  if ((ufd = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
	perror("socket()");
	exit(1);
  }
  memset((void *)&sv_addr, 0, (size_t)sizeof(sv_addr));
  sv_addr.sin_family = AF_INET;
  sv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  sv_addr.sin_port = htons(PORT_NO);
  if (bind(ufd, (struct sockaddr *)&sv_addr, sizeof(sv_addr)) < 0) {
        perror("bind()");
        exit(1);
  }
  if ((argc == 0 || argc == 3) && identify() <= 0)
	error("Can't found Prologix GPIB-ETHERNET CONTROLLER", (char *)0);
  if (argc == 3)
	printf("\n");
  if (argc >= 3 && assign() <= 0)
	error("Can't assign IP address to Prologix GPIB-ETHERNET CONTROLLER",
		(char *)0);
  exit(0);
}

identify(void)
{
  int i, k, n, cadlen;
  char *s;
  unsigned short sum;
  struct sockaddr_in c_addr;
  struct ifreq ifreq;
  unsigned int esum;

  ip.iph.ip_len = htons(40);
  sum = ip_checksum((char *)&ip, 20);
  *(unsigned short *)(&ip.iph.ip_sum) = sum;
  /* UDP header */
  udp_hdr.src_port = htons(PORT_NO);
  udp_hdr.dst_port = htons(PORT_NO);
  udp_hdr.len = htons(20);
  udp_hdr.sum = htons(0);
  /* UDP data */
  nf_identify.hdr.magic = NF_MAGIC;
  nf_identify.hdr.id = NF_IDENTIFY;
  nf_identify.hdr.sequence = htons(rand() & 0xFFFF);
  for (i = 0; i < 6; i++)
	nf_identify.hdr.eth_addr[i] = (uint8_t)0xFF;
  for (i = 0; i < 2; i++)
	nf_identify.hdr.reserved[i] = (uint8_t)0x00;
#ifdef DEBUG
  dump(&nf_identify.hdr, sizeof(struct nf_header)); 
#endif

  memcpy((char *)&ip.data, (char *)&udp_hdr, sizeof(udp_hdr));
  memcpy((char *)&ip.data + 8, (char *)&nf_identify, sizeof(nf_identify));
#ifdef DEBUG
  dump((char *)&ip, 40);
#endif
  memcpy((char *)eth_frame.data, (char *)&ip, 40);
  sum = udp_checksum((char *)&eth_frame.data[0] + 20, 20);
  *(unsigned short *)(&eth_frame.data[0] + 26) = sum;
  /* FCS */
  esum =  eth_checksum((char *)&eth_frame, 54);
  s = (char *)&eth_frame + 58 - 1;
  for (i = 0; i < 4; i++)
        *s-- = (esum >> (i * 8)) & 0xff;
  strcpy(ifreq.ifr_name, ifname);
  if (ioctl(bpf, BIOCSETIF, &ifreq) < 0) {
	perror("ioctl");
	exit(1);
  }
  if ((n = write(bpf, (char *)&eth_frame, 58)) < 0) {
	perror("write");
	exit(1);
  }

  alarm(TIMEOUT);
  for (k = 0;;) {
	cadlen = sizeof(cl_addr);
	if ((n = recvfrom(ufd, buf, sizeof(buf), 0,
			(struct sockaddr *)&cl_addr, &cadlen)) < 0) {
		if (errno == EINTR)
			return k;
		perror("identify(): recvfrom()");
		return -1;
	}
#ifdef DEBUG
  dump(buf, n);
#endif
	if (buf[1] == NF_IDENTIFY_REPLY && !k) {
		if (nf_identify.hdr.sequence != *(unsigned short *)&buf[2])
			continue;
		memcpy(&mac_addr, buf + 4, 6);
		dump_identify_reply((struct nf_identify_reply *)buf);
		k++;
	}
	else if (buf[1] != NF_IDENTIFY_REPLY && buf[1] != NF_IDENTIFY)
		printf("Unexpected Reply: %d\n", buf[1]);
  }
  return -1;
}

assign(void)
{
  int i, k, n, cadlen;
  char *s;
  unsigned short sum, seq;
  struct sockaddr_in c_addr;
  struct ifreq ifreq;
  unsigned int esum;

  ip.iph.ip_len = htons(88);
  sum = ip_checksum((char *)&ip, 20);
  *(unsigned short *)(&ip.iph.ip_sum) = sum;
  /* UDP header */
  udp_hdr.src_port = htons(PORT_NO);
  udp_hdr.dst_port = htons(PORT_NO);
  udp_hdr.len = htons(68);
  udp_hdr.sum = htons(0);
  /* UDP data */
  nf_assignment.hdr.magic = NF_MAGIC;
  nf_assignment.hdr.id = NF_ASSIGNMENT;
  nf_assignment.hdr.sequence = htons(rand() & 0xFFFF);
  for (i = 0; i < 6; i++)
	nf_assignment.hdr.eth_addr[i] = (uint8_t)mac_addr[i];
  for (i = 0; i < 2; i++)
	nf_assignment.hdr.reserved[i] = (uint8_t)0x00;
  for (i = 0; i < 3; i++)
	nf_assignment.reserved[i] = (uint8_t)0x00;
  if (*(unsigned short *)&ip_addr[0] == 0)
	nf_assignment.ip_type = (uint8_t)NF_IP_DYNAMIC;
  else
	nf_assignment.ip_type = (uint8_t)NF_IP_STATIC;
  for (i = 0; i < 4; i++)
	nf_assignment.ip_addr[i] = (uint8_t)ip_addr[i];
  for (i = 0; i < 4; i++)
	nf_assignment.ip_netmask[i] = (uint8_t)ip_netmask[i];

  for (i = 0; i < 4; i++)
	nf_assignment.ip_gw[i] = (uint8_t)ip_gw[i];
  for (i = 0; i < 32; i++)
	nf_assignment.name[i] = (uint8_t)0;
  memcpy((char *)&ip.data, (char *)&udp_hdr, sizeof(udp_hdr));
  memcpy((char *)&ip.data + 8, (char *)&nf_assignment, 60);
  sum = udp_checksum((char *)&ip.data[0], 88);
  *(unsigned char *)&udp_hdr.sum = htons(sum);
#ifdef DEBUG
  dump(&nf_assignment, 88);
#endif
  memcpy((char *)eth_frame.data, (char *)&ip, 88);
  sum = udp_checksum((char *)&eth_frame.data[0] + 20, 20 + 48);
  *(unsigned short *)(&eth_frame.data[0] + 26) = sum;
  /* FCS */
  esum =  eth_checksum((char *)&eth_frame, 102);
  s = (char *)&eth_frame + 106 - 1;
  for (i = 0; i < 4; i++)
	*s-- = (esum >> (i * 8)) & 0xff;
  strcpy(ifreq.ifr_name, ifname);
  if (ioctl(bpf, BIOCSETIF, &ifreq) < 0) {
	perror("ioctl");
	exit(1);
  }
  if ((n = write(bpf, (char *)&eth_frame, 106)) < 0) {
	perror("write");
	exit(1);
  }

  alarm(TIMEOUT);
  for (k = 0;;) {
	cadlen = sizeof(cl_addr);
	if ((n = recvfrom(ufd, buf, sizeof(buf), 0,
			(struct sockaddr *)&cl_addr, &cadlen)) < 0) {
		if (errno == EINTR)
			return k;
		perror("assign(): recvfrom()");
		return -1;
	}
#ifdef DEBUG
  dump(buf, n);
#endif
	if (buf[1] == NF_ASSIGNMENT_REPLY && !k) {
		if (nf_assignment.hdr.sequence != *(unsigned short *)&buf[2])
			continue;
		if (buf[12] != NF_SUCCESS)
			fprintf(stderr,
				"Warning: Network Configuration failed.\n");
		dump_assign_reply((struct nf_assign_reply *)buf);
		k++;
	}
	else if (buf[1] != NF_ASSIGNMENT_REPLY && buf[1] != NF_ASSIGNMENT)
		printf("Unexpected Reply: %d\n", buf[1]);
  }
  return -1;
}

char *
addr(char *p)
{
  static char str[32];
  char *s;
  int i;

  for (s = str, i = 0; i < 4; i++) {
	sprintf(s, "%d", p[i] & 0xFF);
	s += strlen(s);
	if (i < 3) 
		*s++ = '.';
  }
  *s = '\0';
  return str;
}

set_macaddr(char *s)
{
  int i, n, x[6];

  if (strncmp((char *)&mac_addr, "\000\041\151", 3) != 0)
	error("%s is not a Prologix. LLC. device", s);
  n = sscanf(s, "%x:%x:%x:%x:%x:%x", &x[0],  &x[1],
	&x[2], &x[3],  &x[4],  &x[5]);
  if (n != 6)
	error("Illegal Mac Address: %s", s);
  for (i = 0; i < 6; i++)
	mac_addr[i] = x[i];
  return 0;
}

dump_identify_reply(struct nf_identify_reply *p)
{
  int i;

#ifdef DEBUG
  printf("magic: %x\n", (unsigned short)p->hdr.magic);
  printf("id: %x\n", (unsigned short)p->hdr.id);
  printf("sequence: %02x\n", (unsigned short)p->hdr.sequence);
#endif
  printf("Mac address : ");
  for (i = 0; i < 6; i++) {
	if (i > 0)
		printf(":");
	printf("%02x", p->hdr.eth_addr[i]);
  }
  if (strncmp((char *)&mac_addr, "\000\041\151", 3) == 0)
	printf(" (Prologix. LLC.)");
  printf("\n");
  printf(" IP Address : %s (%s)\n", addr((char *)&p->ip_addr),
	p->ip_type ? "Static" : "DHCP");
  printf("   Net Mask : %s\n", addr((char *)&p->ip_netmask));
  printf("    Gateway : %s\n", addr((char *)&p->ip_gw));
  printf("    app_ver : %s\n", addr((char *)&p->app_ver));
  printf("   boot_ver : %s\n", addr((char *)&p->boot_ver));
  printf("     Uptime : %u days %02d:%02d:%02d\n", ntohs(p->uptime_days),
        (int)p->uptime_hrs, (int)p->uptime_mins, (int)p->uptime_secs);
  return 0;
}

dump_assign_reply(struct nf_assignment_reply *p)
{
  int i;

  if (vflg) {
	printf("magic: %x\n", (unsigned short)p->hdr.magic);
	printf("id: %x\n", (unsigned short)p->hdr.id);
	printf("sequence: %02x\n", (unsigned short)p->hdr.sequence);
	printf("Mac address: ");
	for (i = 0; i < 6; i++) {
		if (i > 0)
			printf(":");
		printf("%02x", p->hdr.eth_addr[i]);
	}
	printf("\n");
	printf("result: %x\n", p->result);
  }
  if (p->result == NF_SUCCESS)
	printf("Network Configuration was changed.\n");
  return (p->result == NF_SUCCESS) ? 0 : -1;
}

init(char *ifnam)
{
  char *s, dev[64];
  int fd, i, buflen;
  unsigned short sum;
  struct ifreq ifr;
  struct ifaddrs *ifa, *ifa0;
  struct sockaddr_dl* sdl;
  time_t clk;

  time(&clk);
  srand(clk & 0xFFFF);

  for (i = 0; i < 99; i++) {
	sprintf(dev, "/dev/bpf%i", i );
	if ((bpf = open(dev, O_WRONLY)) != -1)
		break;
  }
  if (i > 99) {
	fprintf(stderr, "Can't open /dev/bpf*\n");
	exit(1);
  }

  /* IP Header */
  ip.iph.ip_v = 4;
  ip.iph.ip_hl = 5;
  ip.iph.ip_tos = 0;
  ip.iph.ip_len = 0x28;
  ip.iph.ip_id = htons(rand() & 0xFFFF);
  ip.iph.ip_off = 0;
  ip.iph.ip_ttl = 0x80; // 255;
  ip.iph.ip_p = IPPROTO_UDP;
  ip.iph.ip_sum = 0;
  if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	perror("socket()");
	exit(1);
  }
  strncpy(ifr.ifr_name, ifnam, sizeof(ifr.ifr_name));

  if (ioctl(fd, SIOCGIFADDR, &ifr) < 0) {
	perror("ioctl()");
	exit(1);
  }
  ip.iph.ip_src.s_addr = *(uint32_t *)&(((struct sockaddr_in *)
	&ifr.ifr_addr)->sin_addr);
  if (ioctl(fd, SIOCGIFBRDADDR, &ifr) < 0) {
	perror("ioctl");
	exit(1);
  }
  ip.iph.ip_dst.s_addr = (uint32_t)0xffffffff;


  if (getifaddrs(&ifa0) < 0) {
	perror("getifaddrs");
	exit(1);
  }
  for (ifa = ifa0; ifa; ifa=ifa->ifa_next) {
	if (ifa->ifa_addr->sa_family != AF_LINK)
		continue;
	if (strcmp(ifa->ifa_name, ifnam) == 0)
		break;
  }
  if (ifa0)
	freeifaddrs(ifa0);
  if (ifa == NULL) {
	fprintf(stderr, "%s not available\n");
	exit(1);
  }
  if (sdl = (struct sockaddr_dl *)ifa->ifa_addr)
	memcpy(if_mac, LLADDR(sdl), sdl->sdl_alen);

  /* build up Ethernet frame header */
  s = &eth_frame.dst[0];
  sscanf("ff:ff:ff:ff:ff:ff", "%x:%x:%x:%x:%x:%x", &s[0], &s[1],
        &s[2], &s[3], &s[4], &s[5]);
  memcpy((char *)&eth_frame.src[0], (char *)&if_mac[0], sizeof(eth_frame.src));
  eth_frame.type[0] = 0x08;
  eth_frame.type[1] = 0x00;

  close(fd);
  return 0;
}

/* find network interface by hostname */

char *
findif(void)
{
  static char ifname[32];
  struct hostent *host;
  in_addr_t haddr;
  struct ifaddrs *ifa, *ifa0;
  struct sockaddr_in *sa;

  if (gethostname(buf, sizeof(buf)) < 0) {
	perror("gethostname()");
	exit(1);
  }
  if ((host = gethostbyname(buf)) == NULL) {
	perror("gethostbyname()");
	exit(1);
  }
  haddr = *(uint32_t *)&host->h_addr[0];
  if (getifaddrs(&ifa0) != 0) {
	perror("get_if_name: getifaddrs() failed");
	exit(1);
  }
  for (ifa = ifa0; ifa; ifa = ifa->ifa_next) {
	sa = (struct sockaddr_in *) ifa->ifa_addr;
	if (sa->sin_addr.s_addr == haddr) {
		strcpy(ifname, ifa->ifa_name);
		break;
	}
  }
  if (ifa0)
	freeifaddrs(ifa0);
  return ifname;
}

/* IP header checksum */

unsigned short ip_checksum(char *s, int n)
{
  register long sum = 0;

  while (n > 1)  {
	sum += *(unsigned short *)s;
	n -= 2;
	s += 2;
  }
  if (n > 0)
	sum += *(unsigned char *)s;
  sum = (sum & 0xffff) + (sum >> 16);
  sum = (sum & 0xffff) + (sum >> 16);
  return ~sum;
}

/* UDP header checksum */

unsigned short udp_checksum(char *s, int n)
{
  register long sum = 0;
  char *t;
  int i;

  /* UDP pseudo header */
  t = (char *)&ip.iph.ip_src;
  for (i = 0; i < 2; i++) {
        sum += *(unsigned short *)t;
        t += 2;
  }
  for (i = 0; i < 2; i++) {
        sum += 0xFFFF;
  }
  sum += htons(*(unsigned char *)&ip.iph.ip_p);
  sum += *(u_int16_t *)&udp_hdr.len;
  /* UDP packet */
  while (n > 1)  {
	if (n > 3)
		sum += *(unsigned short *)s;
	else
		sum += htons(*(unsigned short *)s);
	n -= 2;
	s += 2;
  }
  sum += htons(*(unsigned char *)s);
  sum = ~((sum & 0xffff) + (sum >> 16));
  if (!sum)
	sum = 0xFFFF;
  return sum;
}

/*
 * Ethernet frame FCS calclation
 */

unsigned char rev8(unsigned char c)
{
  unsigned char r = 0;
  int i;

  for (i = 0; i < 8; i++)
	r |= (c & (1 << i)) ? (1 << (7 - i)) : 0;
  return r;
}

unsigned int eth_checksum(char *s, int n)
{
  int i, j;
  unsigned int c, crc0, crc1, fcs;

  crc0 = 0xffffffff;
  for (i = 0; i < n; i++) {
	c = *s++ & 0xFF;
	for (j = 0; j < 8; j++) {
		crc0 = (crc0 << 1) ^ ((((crc0 >> 31) & 1) ^ (c & 1)) ?
			0x04C11DB7 : 0);
	c >>= 1;
    }
  }
  crc1  = rev8(crc0 >> 24) << 24;
  crc1 |= rev8(crc0 >> 16) << 16;
  crc1 |= rev8(crc0 >>  8) <<  8;
  crc1 |= rev8(crc0);
  return (crc1 ^ 0xFFFFFFFF);
}

set_arg(char *s, unsigned char a[4])
{
  int i;
  char *t;

  for (t = s, i = 0; i < 4; i++) {
	for (t = s; *t && *t != '.'; t++)
		;
	a[i] = (unsigned char)atoi(s);
	s = t;
	if (*s == '.')
		s++;
  }
  if (i < 4)
	fprintf(stderr, "Illegal IP address\n");
  return 0;
}

#ifdef DEBUG
dump(char *s, int n)
{
  int i;

  for (i = 0; i < n; i++) {
	printf(" %02x", s[i] & 0xFF);
	if ((i % 16) == 15)
		printf("\n");
  }
  printf("\n");
}
#endif

void
onalarm(int sig)
{
  alarm(0);
  saalrm.sa_handler = onalarm;
  sigemptyset(&saalrm.sa_mask);
  saalrm.sa_flags = 0;
  sigaction(SIGALRM, &saalrm, &osaalrm);
}

error(char *s, char *t)
{
  fprintf(stderr, s, t);
  fprintf(stderr, "\n");
  exit(1);
}

Usage(void)
{
  static char *t[] = {
"Command   %s - PROLOGIX GPIB-ETHERNET CONTROLLER Network Configurator",
"Syntax    %s [ option ] [ ip_address network_mask gateway [ MAC_address ]]",
"Option    -i interface	# select network interface",
"Example   %s		# find PROLOGIX GPIB-ETHERNET CONTROLLER",
"          %s 192.168.1.106 255.255.255.0 192.168.1.1	# configure",
"          %s -i re0 192.168.1.106 255.255.255.0 192.168.1.1",
"          %s 192.168.1.106 255.255.255.0 192.168.1.1 00:21:69:01:02:4",
"          %s 0.0.0.0 255.255.255.0 192.168.1.1	# DHCP setup",
"If invoked with no arguments, this progaram try to find PROLOGIX",
"GPIB-ETHERNET CONTROLLER.  If invoked with three araguments, this program",
"assin specified network parameters to found PROLOGIX GPIB-ETHERNET",
"CONTROLLER.  If invoked with four araguments, this program assin specified",
"network parameters to GPIB-ETHERNET CONTROLLER which has specified MAC",
"address which will be obtained from the device lable atatched on back of",
"device box.  If ip_address is specifies as 0.0.0.0, GPIB-ETHERNET CONTROLLER",
"is configured to use DHCP.  On multi-homed hosts, select network interface by",
"-i option.",
"Bug	Only super user can execute this command.  This program only works on",
"FreeBSD which has bpf(4) devices.  There may be some little endian dipendent",
"codes.",
  NULL
  };
  int i;

  for (i = 0; t[i] != NULL; i++) {
	printf(t[i], cmd);
	printf("\n");
  }
  exit(1);
}

